/* $Id: grid.h,v 1.4 2006/01/18 12:00:11 cvsroot Exp $ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GENLIB2_GRID_INCLUDED
#define GENLIB2_GRID_INCLUDED

#define MAXGRID 128

using namespace std;

namespace GenLib2
{

/// 2D point
struct point {double x, y; int c;};

/// 2D rectangle
struct rect {double x1, y1, x2, y2;};

/// range
struct range {double min, max;};


/** Grid method for range searching.
 *  Simple method for maintaining proximity relationship among
 *  points in a plane. An artificial grid divided an defined
 *  area in small rectagular cells. The points are
 *  keep in lists of points falling into each cell.
 *
 *  Reference: Robert Sedgewick, Algorithms in C++,
 *  Addision-Wesley Publishing Company, 1992.
 *
 *  @author Dirk Stallmann
 */
class Grid
{
public:
	/// constructor
	Grid();

	/// destructor
	~Grid();

	/// clear grid, erase all points
	void clear();

	/// set gird resolution
	void set(struct range x, struct range y);

	/// show infomation
	void info(ostream& os) const;

	/// insert point
	void insert(struct point p);

	/// count number of points
	size_t size() const;

	/// print number of points per grid square
	void occupied(ostream& os) const;

	/** calculate average inside the given rectangle.
	 *  by weighted distance (p=1/s^2)
	 *  distance s = |dx| + |dy|
	 */
	int average(struct rect range, struct point& p) const;

	/// find bonding box
	bool bbox(struct range& rx, struct range& ry) const;

	/// find closest point
	struct point* closest(struct rect range, struct point& o) const;

	/// print grid points
	void print(ostream& os) const;

	/// print grid points using format string
	void print(ostream& os, char* fmt) const;

protected:
	/// node definition
	struct node
		{
		/// point
		struct point p;

		/// next node
		struct node* next;
		};

	/// grid array width fixed size
	struct node* grid[MAXGRID][MAXGRID];

	/// dummy tail node
	struct node* z;

	/// origin
	struct point origin;

	/// grid width in x and y
	double width;

private:
	/// squared minimal distance
	double min2;

	/// closest points
	struct point cp1, cp2;

	/// compute distance and check if less than min
	void check(struct point p1, struct point p2);

	/// check if inside rectangle
	int insiderect(struct point p, struct rect range) const;
};

} // namespace GenLib2

#endif
