/* $Id: matrix.h,v 1.7 2007/12/10 09:04:55 cvsroot Exp $ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GENLIB2_MATRIX_INCLUDED
#define GENLIB2_MATRIX_INCLUDED

#include <cassert>
#include <iostream>
#include <cmath>
using namespace std;

// forward declarations
//template<class T> class IPL::Image;

namespace GenLib2
{

/** Matrix class.
 *  General matrix class.
 *  @author Dirk Stallmann
 */
template <class T>
class Matrix
{
	//friend ostream& operator<<(ostream& os, const Matrix<T>& m);
	//friend class Image<T>;
	
protected:
	/**@name internal helper functions.
	 */
	//@{
		/// allocate memory
		void allocate(int m, int n);

		/// free memory
		void free();

		/// copy array
		void copy(const T* v);

		/// set all elements to value
		void set(const T& val);
	//@}
	
	//T* data() const { return _data; }
	//T* row() const { return _row; }

public:
	/**@name constructors.
	 */
	//@{
		/// constructor without size
		Matrix();

		/// constructor with size and initilize with value
		Matrix(int m, int n, const T& val=T(0));

		/// constructor, initialize with array
		Matrix(int m, int n, const T* v);

		/// constructor of row vector and initilize with zeros
		Matrix(int m, const T& val=T(0));

		/// constructor of row vector, initialize with array
		Matrix(int m, const T* v);

		// with argument list
		//Matrix(int m, int n, T val0, ...);
	
		/// copy constructor
		Matrix(const Matrix& A);
	//@}
	
	/// destructor
	~Matrix();

	/**@name methods.
	 */
	//@{ 
		/// copy operator
		Matrix& operator=(const Matrix& A);

		/// copy sub-matrix A into *this matrix
		Matrix& copy(int rl, int cl, const Matrix& A);

		/// new size
		Matrix& newsize(int m, int n=1);

		/// set all elements to zero
		Matrix& clear() { set(T(0)); return *this; }

		/// set all elements to zero
		void null();

		/// set diagonal elements to 1 (and don't change off-diagonal elements)
		void set_diagonal();

		/** set identity matrix.
		 *  Set diagonal elements to value and non-diagonal elements
		 *  will be cleared.
		 */
		void identity(T val=T(1));

		/// set all elements to value
		Matrix& set_all(const T& val);

		// get and set rows and columns
		// gets and sets rows and columns in (*this). The supplied
		// row/column vector must have exactly the right number of
		// elements. It is not important whether it is an row (1xN)
		// or column (Nx1) vector.
		//void getRow(int rowIndex, Matrix &row);
		void get_col(int c, Matrix& col);
		//void setRow(int rowIndex, const Matrix &row);
		//void setCol(int colIndex, const Matrix &col);

		/// make a square matrix symmetric (by mirroring the upper part)
		void make_upper_symmetric();

		/// make a square matrix symmetric (by mirroring the lower part)
		void make_lower_symmetric();
	//@}
	
	/**@name member data access.
	 */
	//@{
		/// number of rows
		int rows() const { return _m; }
		
		/// number of columns
		int cols() const { return _n; }
		
		/// number of matrix elements
		int size() const { return _m * _n; }

		/// data check
		bool is_data() const { return (_data != NULL); }
		
		/// data check
		bool is_row() const { return (_row != NULL); }

		/// get matrix element function using (i,j) indexing
		T& elem(int i, int j) const { return _row[i][j]; }
		
		/// set matrix element function using (i,j) indexing
		void elem(int i, int j, T val) { _row[i][j] = val; }

		/// get matrix element using (i,j) indexing
		T& operator()(int i, int j) { return _row[i][j]; }
		
		/// get matrix element using (i,j) indexing
		const T& operator()(int i, int j) const { return _row[i][j]; }

		/// get matrix element of row-vector using (i) indexing
		T& operator()(int i) { return _row[i][0]; }
		
		/// get matrix element of row-vector using (i) indexing
		const T& operator()(int i) const { return _row[i][0]; }

		/// get pointer to a row
		T* row(int i) const { return _row[i]; }
		// begin of data
		//T* beg() { return _data[0]; }
		// end of data
		//T* end() { return _data[0] + _m * _n; }

		//int inside(int r, int c) {			// check if inside matrix
		//	return (r >= 0 && c >= 0 && r < _m && c < _n);
		//}
	//@}
	
	/**@name I/O methods.
	 */
	//@{
		/// read matrix
		void read(std::istream& is);
	
		/// print matrix
		void print(ostream& os) const;

		/// print symbolic
		void print_symb(ostream& os, char zero=' ', char nozero='*') const;

		/// print in PBM-format
		void print_pbm(ostream& os) const;
	//@}
	
	/**@name composite assignment operators.
	 */
	//@{
		/// multiplication with a scalar
		Matrix& operator*=(T s);

		/// division by a scalar
		Matrix& operator/=(T s);
	//@}
	
	/**@name 
	 * these are routines taken from Numerical Recipes in C.
	 * They are only to be used for
	 * REAL, SYMMETRIC matrices (square matrices, of course).
	 */
	 //@{
		/** householder reduction
		 * (numerical recipes, page 373)
		 * householder reduction (to a tridiagonal matrix) for
		 * a real, symmetric matrix
		 * (*this) is replaced by the orthogonal matrix Q
		 *  d returns the diagonal elements
		 *  e returns the off-diagonal elements with e(0)=0
		 */
		void tred2(Matrix& d, Matrix& e);

		/** QL algorithm with implicit shifts
		 * (numerical recipes, page 380)
		 *  d and e are the elements of a real, symmetric, tridiagonal
		 *     matrix (d=diagonal, e=off-diagonal elements, e(0) arbitrary).
		 *  (*this) is either the identity matrix in which case the
		 *    eigenvectors of the tridiagonal matrix are returned or
		 *     the matrix as returned by tred2. Then, the eigenvectors
		 *    of the original 'A'-matrix are returned.
		 *  maxIter is the maximum allowed number of iterations (if present)
		 *
		 * RETURNS:
		 *  d(i)   : eigenvalue i
		 *  e      : destroyed
		 *  (*this): k-th column contains normalized eigenvector
		 *           corresponding to d(i)
		 *  All results are only valid if the routine returns TRUE
		 */
		bool tqli(Matrix& d, Matrix& e, int maxIter=30);

		/**
		 * Determines eigenvalues and eigenvectors f the (*this) matrix
		 * for a real, symmetric matrix
		 *  RETURNS:
		 * (the value of *this is preserved)
		 * eigenVals: Nx1 vector of all eigenvalues
		 * eigenVects: NxN matrix of all eigenvectors, written as columns
		 * (i-th column eigenVects(*,i) corresponds to eigenvalue eigenVals(i))
		 *
		 * Result only valid if the routine returns TRUE
		 */
		bool eigensystem(Matrix& eigen_vals, Matrix& eigen_vecs);
	//@}
	
protected:
	/// matrix dimension, number of rows
	int		_m;
	
	/// matrix dimension, number of columns
	int		_n;

	/// matrix elements
	T*		_data;

	/// row-pointer
	T**		_row;
};


// non-member functions...

// I/O operators

/// input operator
template <class T>
istream& operator>>(istream& is, Matrix<T>& m);

/// output operator
template <class T>
ostream& operator<<(ostream& os, const Matrix<T>& m);


// basic matrix algorithms

/// transpose matrix
template <class T>
Matrix<T> transpose(const Matrix<T>& A);

// add/sub matrix m to/from the *this matrix
//Matrix &add(const Matrix &m);
//Matrix &sub(const Matrix &m);

/// addition operator
template <class T>
Matrix<T> operator+(const Matrix<T>& A, const Matrix<T>& B);

/// subtraction operator
template<class T>
Matrix<T> operator-(const Matrix<T>& A, const Matrix<T>& B);

/// matrix multiplication function using references
template <class T>
void mult(Matrix<T>& C, const Matrix<T>& A, const Matrix<T>& B, bool clr=true);

/// matrix multiplication function
template <class T>
Matrix<T> mult(const Matrix<T>& A, const Matrix<T>& B);

/// matrix multiplication operator
template <class T>
Matrix<T> operator*(const Matrix<T>& A, const Matrix<T>& B);

/// general matrix multipication function C = A * B
// matrix status
//   clr=false ... Matrix C accummulated
//   clr=true  ... Matrix C cleared in advance
//   atr=false ... matrix A as defined
//   atr=true  ... matrix A transposed
//   btr=false ... matrix B as defined
//   btr=true  ... matrix B transposed
template <class T>
void mult_gen(Matrix<T>& C, bool clr, const Matrix<T>& A, bool atr,
	const Matrix<T>& B, bool btr);

/// multiplication with scalar
template <class T>
Matrix<T> operator*(double s, const Matrix<T>& A);


// implementation

// allocate memory

template <class T>
void Matrix<T>::allocate(int m, int n)
{
	assert(_data == NULL); // to avoid memory leaks

	_m = m;
	_n = n;

	// allocate matrix elements
	_data = (T*)new T[_m * _n];
	assert(_data != NULL);

	// allocate array of pointers to rows (row-pointers)
	_row = new T*[_m];
	assert(_row != NULL);

	// set row-pointers
	_row[0] = _data;
	for (register int i = 1; i < _m; i++)
		_row[i] = _row[i-1] + _n;
}


// free data

template <class T>
void Matrix<T>::free()
{
	// do nothing, if no memory has been previously allocated
	if (_data == NULL) return;

	// free previously allocated matrix
	if (_data != NULL) delete[] _data;

	// free row-pointers
	if (_row != NULL) delete[] _row;

	_data = NULL;
}


// copy array

template <class T>
void Matrix<T>::copy(const T* v)
{
	assert(_data != NULL);
	for (int i = 0; i < _m * _n; i++)
		_data[i] = v[i];
}


// set all elements to value

template <class T>
void Matrix<T>::set(const T& val)
{
	//assert(_data != NULL);
	if (_data == NULL) return; // nothing to do!
	for (int i = 0; i < _m * _n; i++)
		_data[i] = val;
}


// constructor without size (default)

template <class T>
Matrix<T>::Matrix() : _m(0), _n(0), _data(NULL), _row(NULL) {}


// constructor with size and initilize with value

template <class T>
Matrix<T>::Matrix(int m, int n, const T& val) : _m(0), _n(0),
_data(NULL), _row(NULL)
{
	allocate(m, n);
	set(val);
}


// constructor, initialize with array

template <class T>
Matrix<T>::Matrix(int m, int n, const T* v) : _m(0), _n(0),
_data(NULL), _row(NULL)
{
	allocate(m, n);
	copy(v);
	//memcpy(_data[0], val, m * n * sizeof(T));
}


// constructor of row vector, initilize with value

template <class T>
Matrix<T>::Matrix(int m, const T& val) : _m(0), _n(0),
_data(NULL), _row(NULL)
{
	allocate(m, 1);
	set(val);
}


// constructor of row vector, initialize with array

template <class T>
Matrix<T>::Matrix(int m, const T* v) : _m(0), _n(0),
_data(NULL), _row(NULL)
{
	allocate(m, 1);
	copy(v);
	//memcpy(_data[0], val, m * sizeof(T));
}


// constructor, initialize with argument list
// Using size_y instead of val0 to cirrcumvent problems
// with variable argument lists
// Attention: This function doesn't work with float!
/*
template <class T>
Matrix<T>::Matrix(int m, int n, T val0, ...)
{
	DB_ENTER("Matrix::arg list constr");
	va_list ap;
	int i;

	allocate(m, n);
	int size = m * n;
	T* p = _data[0];
	va_start(ap, n);
	*p = val0;
	for (i = 0; i < size; i++)
		(*(p++) = va_arg(ap, T));
	va_end(ap);
}
*/


// copy constructor

template <class T>
Matrix<T>::Matrix(const Matrix& A) : _m(0), _n(0),
_data(NULL), _row(NULL)
{
	allocate(A._m, A._n);
	copy(A._data);
}


// destructor

template <class T>
Matrix<T>::~Matrix()
{
	free();
}


// new matrix size
// set matrix dimensions (redimension only when required)

template <class T>
Matrix<T>& Matrix<T>::newsize(int m, int n)
{
	if (_m == m && _n == n) return *this;
	free();
	allocate(m, n);
	return *this;
}


// assignment, copy data

template <class T>
Matrix<T>& Matrix<T>::operator=(const Matrix& A)
{
	if (_data == A._data) return *this;
	
	if (_m == A._m && _n == A._n)
		copy(A._data);
	else
		{
		newsize(A._m, A._n);
		copy(A._data);
		//memcpy((void *)_data[0], (void *)m._data[0],
		//	_m * _n * sizeof(T));
		}

	return *this;
}


// copy sub-matrix A into *this matrix

template <class T>
Matrix<T>& Matrix<T>::copy(int rl, int cl, const Matrix<T>& A)
{
	// check range!!!!
	assert(0 <= rl);
	assert(rl < _m);
	assert(0 <= cl);
	assert(cl < _n);

	assert(rl+A._m < _m);
	assert(cl+A._n < _n);

	for (int i = 0; i < A._m; i++)
		for (int j = 0; j < A._n; j++)
			_row[rl+i][cl+j] = A._row[i][j];

	return *this;
}


// set all elements to zero
/*
template <class T>
Matrix<T>& Matrix<T>::clear()
{
	set(T(0));
	//memset((void *)_data[0], 0, _m * _n * sizeof(T));

	return *this;
}
*/


// set all elements to zero

template <class T>
void  Matrix<T>::null()
{
	set(T(0));
}


// set diagonal elements to 1 (and don't change off-diagonal elements)

template <class T>
void Matrix<T>::set_diagonal()
{
	register int i = _n < _m ? _n : _m;

	for(i -= 1; i >= 0; --i)
		_row[i][i] = T(1);
}


// set identity matrix

template <class T>
void Matrix<T>::identity(T val)
{
	assert(_n == _m);

	set(T(0));
	for(int i = 0; i < _m; i++)
		_row[i][i] = val;
}


// set all elements to value

template <class T>
Matrix<T>& Matrix<T>::set_all(const T& val)
{
	set(val);
	return *this;
}


// get and set rows and columns
/*
template <class ElType>
void Matrix<ElType>::getRow(int rowIndex, Matrix<ElType> &row) {
    register int i;
    ElType *src, *dest;
    assert(colNum==(row.rows()*row.cols()));

    src=data+rowIndex*colNum;
    dest=row.data;
    for(i=colNum; i>0; --i)
    	*dest++ = *src++;
    }
*/

template <class T>
void Matrix<T>::get_col(int c, Matrix<T>& col)
{
	col.newsize(_n);

	for (int i = 0; i < _n; i++)
		col._row[i][0] = _row[i][c];
}

/*
template <class ElType>
void Matrix<ElType>::setRow(int rowIndex, const Matrix<ElType> &row) {
    register int i;
    ElType *src, *dest;
    assert(colNum==(row.rows()*row.cols())); 

    src=row.data;
    dest=data+rowIndex*colNum;
    for(i=colNum; i>0; --i)
    	*dest++ = *src++;
    }

template <class ElType>
void Matrix<ElType>::setCol(int colIndex, const Matrix &col) {
    register int i;
    ElType *src, *dest;
    assert(rowNum==(col.rows()*col.cols()));

    src = col.data;
    dest = data+colIndex;
    for(i=rowNum; i>0; --i) {
	*dest = *src++;
	dest += colNum;
    	}
    }
*/


// make a square matrix symmetric (by mirroring the upper part)

template <class T>
void Matrix<T>::make_upper_symmetric()
{
	//register T *pr, *pc;
	register int r, c;

    // only useful for square matrices
    assert(_m == _n);

    for(r = 0; r < _m; r++)
		for(c = r; c < _n; c++)
			_row[c][r] = _row[r][c];
/*
    for(r = 0; r < _m; ++r)
		{
		//pr = pc = _data[0] + r * (_n+1);
		pr = pc = &_data[r][r];
		for(c = r; ++c < _n; )
			{
			++pr;
			pc += _n;
			*pc = *pr;
			}
		}
*/
}


// make a square matrix symmetric (by mirroring the lower part)

template <class T>
void Matrix<T>::make_lower_symmetric()
{
	//register T *pr, *pc;
	register int r, c;

	// only useful for square matrices
	assert(_m == _n);

    for(r = 0; r < _m; r++)
		for(c = r; c < _n; c++)
			_row[r][c] = _row[c][r];
/*
	for(r = 0; r < _m; ++r)
		{
		//pr = pc = _data[0]+ r*(_n+1);
		pr = pc = &_data[r][r];
		for(c = r; ++c < _n; )
			{
			++pr;
			pc += _n;
			*pr = *pc; // this line makes the difference...
			}
		}
*/
}


// read data

template <class T>
void Matrix<T>::read(istream& is)
{
	int m = 0, n = 0;
	is >> m >> n;
	newsize(m, n);

	for (int i = 0; i < m; i++)
		for (int j = 0; j < n; j++)
			is >> _row[i][j];
}


// print matrix

template <class T>
void Matrix<T>::print(ostream& os) const
{
	os << _m << " " << _n << endl;
	for (int i = 0; i < _m; i++)
		{
		for (int j = 0; j < _n; j++)
			os << _row[i][j] << " ";
		os << endl;
		}
}


// print symbolic

template <class T>
void Matrix<T>::print_symb(ostream& os, char zero, char nozero) const
{
	int r, c;

	// print the matrix size first
	os << _m << " x " << _n << ":" << endl;

	//T* p = _data[0];
	for (r = 0; r < _m; r++)
		{
    	for(c = 0; c < _n; c++)
	    //if (*p++ == (T)0)
	    if (_row[r][c] == (T)0)
			os << zero;
		else
			os << nozero;
		os << endl;
		}
}


// print in PBM-format

template <class T>
void Matrix<T>::print_pbm(ostream& os) const
{
	int r, c;

	// PBM header
	os << "P1\n" << _n << " " << _m << "\n" << endl;

	// print the data
	//T* p = _data[0];
	for (r = 0; r < _m; r++)
		{
    	for(c = 0; c < _n; c++)
	    //if (*p++ == (T)0)
	    if (_row[r][c] == (T)0)
			os << "0 ";
		else
			os << "1 ";
		os << endl;
		}
}


// multiplication with a scalar

template <class T>
Matrix<T>& Matrix<T>::operator*=(T s)
{
	for (int i = 0; i < _n; i++)
		for (int j = 0; j < _m; j++)
			_data[i][j] *= s;
		return *this;
}


// division by a scalar

template <class T>
Matrix<T>& Matrix<T>::operator/=(T s)
{
	for (int i = 0; i < _n; i++)
		for (int j = 0; j < _m; j++)
			_data[i][j] /= s;
		return *this;
}


// tred2
// householder reduction (to a tridiagonal matrix) for
// a real, symmetric matrix
template <class T>
void Matrix<T>::tred2(Matrix<T>& d, Matrix<T>& e)
{
	int l, k, j, i;
	T scale, hh, h, g, f;

	// dimension checks...
	assert((_n == _m) && (d.rows() >= _n) && (e.rows() >= _n));

	// alg from numerical recipes
	for (i = _n - 1; i >= 1; --i)
		{
		l = i - 1;
		h = scale = 0.0;
		if (l > 0)
			{
			for (k = 0; k <= l; ++k) scale += (T)fabs((*this)(i,k));
			if(scale == 0.0)
				e(i) = (*this)(i,l);
			else
				{
				for (k = 0; k <= l; ++k)
					{
					T tmp;
					tmp = ((*this)(i,k) /= scale);
					h += tmp * tmp;
					}
				f = (*this)(i,l);
				g = (f > 0) ? -(T)sqrt((double)h) : (T)sqrt((double)h);
				e(i) = scale * g;
				h -= f * g;
				(*this)(i,l) = f - g;
				f = 0.0;
				for (j = 0; j <= l; ++j)
					{
					(*this)(j,i)=(*this)(i,j) / h;
					g = 0.0;
					for (k = 0; k <= j; ++k) g += (*this)(j,k) * (*this)(i,k);
					for (k = j + 1; k <= l; ++k) g += (*this)(k,j) * (*this)(i,k);
					e(j) = g / h;
					f += e(j) * (*this)(i,j);
					}
				hh = f / (h + h);
				for(j = 0; j <= l; ++j)
					{
					f = (*this)(i,j);
					g = (e(j) -= hh * f);
					for (k = 0; k <= j; ++k) (*this)(j,k) -= (f * e(k) + g * (*this)(i,k));
					}
				}
			}
		else
			e(i) = (*this)(i,l);
			
		d(i) = h;
		}

		d(0) = e(0) = (T)0.0;
		for(i=0; i<_n; ++i)
			{
			l = i - 1;
			if (d(i) != 0.0)
				{
				for (j = 0; j <= l; ++j)
					{
					g = 0.0;
					for(k=0; k<=l; ++k) g += (*this)(i,k) * (*this)(k,j);
					for(k=0; k<=l; ++k) (*this)(k,j) -= g * (*this)(k,i);
					}
				}
			d(i) = (*this)(i,i);
			(*this)(i,i) = (T)1.0;
			for (j = 0; j <= l; j++) (*this)(j,i) = (*this)(i,j) = 0.0;
		}
}


// QL algorithm with implicit shifts

template <class T>
bool Matrix<T>::tqli(Matrix<T>& d, Matrix<T>& e, int maxIter)
{
	int m, l, iter, i, k;
	T s, r, p, g, f, dd, c, b;

	// simple check...
	assert(_n == _m && _n == d.rows() && _n == e.rows());

	// alg from numerical recipes
	for (i = 1; i < _n; ++i) e(i-1) = e(i);
	e(_n-1) = (T)0.0;
	for (l = 0; l < _n; ++l)
		{
		iter = 0;
		do
			{
			for (m = l; m < _n - 1; ++m)
				{
				dd = fabs(d(m)) + fabs(d(m+1));
				if ((T)(fabs(e(m)) + dd) == dd) break;
				}
			if (m != l)
				{
				if (iter++ == maxIter)
					{
					cerr << "Matrix<...>::tqli: maxIter exceeded" << endl;
					return false;
					}
				g = (d(l+1) - d(l)) / (2.0 * e(l));
				r = (T)sqrt((g * g) + 1.0);
				g = (T)(d(m) - d(l) + e(l) /(g + ((g < 0.0)? -fabs(r) : fabs(r))));
				s = c = (T)1.0;
				p = (T)0.0;
				for (i = m - 1; i >= l; --i)
					{
					f = s * e(i);
					b = c * e(i);
					if (fabs(f) >= fabs(g))
						{
						c = g / f;
						r = (T)sqrt(c * c + 1.0);
						e(i+1) = f * r;
						c *= (s = 1.0 / r);
						}
					else
						{
						s = f / g;
						r = (T)sqrt(s * s + 1.0);
						e(i+1) = g * r;
						s *= (c = 1.0 / r);
						}
					g = d(i+1) - p;
					r = (d(i) - g) * s + 2.0 * c * b;
					p = s * r;
					d(i+1) = g + p;
					g = c * r - b;
					for (k = 0; k < _n; ++k)
						{
						f = (*this)(k,i+1);
						(*this)(k,i+1) = s * (*this)(k,i) + c * f;
						(*this)(k,i) = c * (*this)(k,i) - s * f;
						}
					}
				d(l) = d(l) - p;
				e(l) = g;
				e(m) = (T)0.;
				}
			}
			while (m != l);
		}

    return true;
}


// determines eigenvalues and eigenvectors 

template <class T>
bool Matrix<T>::eigensystem(Matrix<T> &eigenVals, Matrix<T> &eigenVects)
{
	Matrix<T> e(_n);

	// should test for symmetric matrix as well...
	assert(_n == _m
		&& eigenVals.rows() == _n && eigenVals.cols() == 1
		&& eigenVects.rows() == _n && eigenVects.cols() == _n);

	// reduce to tridiagonal matrix
	eigenVects = (*this);
	eigenVects.tred2(eigenVals, e);

	// get eigenvalues and eigenvectors
	return(eigenVects.tqli(eigenVals, e));
}


// non-member functions...

// I/O operators

template <class T>
istream& operator>>(istream& is, Matrix<T>& m)
{
	m.read(is);
	return is;
}

template <class T>
ostream& operator<<(ostream& os, const Matrix<T>& m)
{
	m.print(os);
	return os;
}


// basic matrix algorithms

// Using the "return value language extension",
// see Lippman&Lajoie: C++ Primer, 3rd edition, p. 735

// transpose function

template <class T>
Matrix<T> transpose(const Matrix<T>& A)
{
	int m = A.rows();
	int n = A.cols();

    Matrix<T> B(n, m);
	
	for (register int i = 0; i < m; i++)
		for (register int j = 0; j < n; j++)
			B(j,i) = A(i,j);

    return B;
}


// add matrix m to the *this matrix
/*
template<class T>
Matrix<T>& Matrix<T>::add(const Matrix<T> &m)
{
	DB_ENTER("Matrix::add to this");

	register T *src = _data[0], *dest = m._data[0];
	assert(_m == m._m && _n == m._n);

	for (register int i = size(); i > 0; --i)
		*dest++ += *src++;

	return (*this);
}
*/


// sub matrix m from the *this matrix
/*
template<class T>
Matrix<T>& Matrix<T>::sub(const Matrix<T> &m)
{
	DB_ENTER("Matrix::sub from this");

	register T *src = _data[0], *dest = m._data[0];
	assert(_m == m._m && _n == m._n);

	for (register int i = size(); i > 0; --i)
		*dest++ -= *src++;

	return (*this);
}
*/


// addition operator

template <class T>
Matrix<T> operator+(const Matrix<T>& A, const Matrix<T>& B)
{
	int m = A.rows();
	int n = A.cols();

	assert(m == B.rows());
	assert(n == B.cols());

    Matrix<T> C(m, n);
	
	for (register int i = 0; i < m; i++)
		for (register int j = 0; j < n; j++)
			C(i,j) = A(i,j) + B(i,j);

	return C;
}


// subtraction operator

template <class T>
Matrix<T> operator-(const Matrix<T>& A, const Matrix<T>& B)
{
	int m = A.rows();
	int n = A.cols();

	assert(m == B.rows());
	assert(n == B.cols());

    Matrix<T> C(m, n);
	
	for (register int i = 0; i < m; i++)
		for (register int j = 0; j < n; j++)
			C(i,j) = A(i,j) - B(i,j);

	return C;
}


// matrix multiplication function using references
// Maybe obsolte by using the "return value language extension".

template <class T>
void mult(Matrix<T>& C, const Matrix<T>& A, const Matrix<T>& B, bool clr)
{
	int m = A.rows();
	int n = A.cols();
	int o = B.rows();
	int p = B.cols();

    // check dimensions
    assert(n == o);

	// set matrix dimensions (redimension only when required)
    C.newsize(m, p);

	// clear in advance
	if (clr) C.clear();
/*
    // trying to get multiplication fast...
    register int i, j, k;
    register T *pc = _data[0];

    for (i = 0; i < _m; ++i)
		for (j = 0; j < _n; ++j)
			{
			register T sum = (T)0;
			register T* pa = a._data[0] + i * a._n;
			register T* pb = b._data[0] + j;

			for	(k = 0; k < a._n; ++k)
				{
				sum += *pa++ * *pb;
				pb += b._n;
				}
	
		    *pc++ = sum;
		    }
*/
    for (register int i = 0; i < m; ++i)
		for (register int j = 0; j < p; ++j)
			{
			register T sum = (T)0;

			for	(register int k = 0; k < n; ++k)
				{
				sum += A(i,k) * B(k,j);
				}
	
		    C(i,j) += sum;
		    }
}


// matrix multiplication function

template <class T>
Matrix<T> mult(const Matrix<T>& A, const Matrix<T>& B)
{
    Matrix<T> C;
	mult(C, A, B, true);
	return C;
}


// matrix multiplication operator

template <class T>
Matrix<T> operator*(const Matrix<T>& A, const Matrix<T>& B)
{
    return mult(A,B);
}


// general multiplication function

template <class T>
void mult_gen(Matrix<T>& C, bool clr, const Matrix<T>& A, bool atr,
const Matrix<T>& B, bool btr)
{
    // check dimensions
	int pa = atr ? A.rows() : A.cols();
	int pb = btr ? B.cols() : B.rows();
    assert(pa == pb);

	// number of rows and columns
	int n = atr ? A.cols() : A.rows();
	int m = btr ? B.rows() : B.cols();

	// set/adjust dimensions of the result matrix
	C.newsize(n, m);

	// clear in advance
	if (clr) C.clear();

    // trying to get multiplication simple...
    for (register int i = 0; i < n; ++i)
		for (register int j = 0; j < m; ++j)
			{
			register T sum = (T)0;

			for	(register int k = 0; k < pa; ++k)
				{
				register T av = atr ? A(k,i) : A(i,k);
				register T bv = btr ? B(j,k) : B(k,j);
				sum += av * bv;
				}
	
		    C(i,j) += sum;
			}
}


// multiplication with a scalar

template <class T>
Matrix<T> operator*(double s, const Matrix<T>& A)
{
	int m = A.rows();
	int n = A.cols();

    Matrix<T> C(m, n);
	
	for (register int i = 0; i < m; i++)
		for (register int j = 0; j < n; j++)
			C(i,j) = s * A(i,j);

	return C;
}

} // namespace GenLib2

#endif
