/* $Id: rotmatrix3.h,v 1.16 2006/07/07 11:55:53 cvsroot Exp $ */

/* GenLib2 - General Library 2
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef GENLIB2_ROTMATRIX3_INCLUDED
#define GENLIB2_ROTMATRIX3_INCLUDED

#include <iostream>
#include <cassert>
#include <vector>
//#include <cmath>  Does't work on SGI because the math include file does't
// fit to the compiler GNU gcc 2.95.2.
// Error message: /usr/include/math.h:855: redefinition of `class exception'
// For a simple work around the forward declarations of required C functions
// are used (see below).

using namespace std;

#include "rmat3supp.h"

// For GNU gcc explicitly state that the friend function is a template,
// by appending angle brackets to its name. MS VC++ do not understand the
// angle brackets.
#if defined(__GNUG__)
#define FRIEND_FUNC_TEMPLATE <>
#else
#define FRIEND_FUNC_TEMPLATE
#endif

// some simple macros because anonymous unions does't work width g++ 2.95.2
#define A11 _a[0][0]
#define A12 _a[0][1]
#define A13 _a[0][2]
#define A21 _a[1][0]
#define A22 _a[1][1]
#define A23 _a[1][2]
#define A31 _a[2][0]
#define A32 _a[2][1]
#define A33 _a[2][2]

// forward declarations of mathematical functions
extern "C" {
	double sqrt(double);
	double sin(double);
	double cos(double);
	double asin(double);
	double atan2(double, double);
}

namespace GenLib2
{
// some forward declarations
template <class T>
class RotMatrix3;

template <class T>
istream& operator>>(istream& is, RotMatrix3<T>& M);
template <class T>
ostream& operator<<(ostream& os, const RotMatrix3<T>& M);
template <class T>
RotMatrix3<T> operator*(const RotMatrix3<T>& A, const RotMatrix3<T>& B);


/** 3x3 rotation matrix class.
 *  Template class to rotate a 3D coordinate system.
 *
 *  The lineare transformation formular can be written by means of
 *  matrix/vector algebra:
 *
 *    r' = A r
 *
 *  where
 *    r    coordinate vector of point P
 *    r'   coordinate vector of point P'
 *    A    the orthogonal rotation matrix.
 *
 *  The rotation matrix is represented by a 3x3 orthogonal matrix.
 *  The matrix elements are the cosines of the space angles between
 *  corresponding axes of the two coordinate systems, also called
 *  'direction cosines'. 
 *
 *  Two general remarks:
 *  (a) transformation direction
 *  transformation direction of the rotation matrix,
 *  when matrix A transform from coordinate system K into K'
 *  the transposed matrix A^t transform system K' into K. 
 *  
 *  (b) active/passive transformation
 *  In general we have to distinguish between the active (b) and the passive (a)
 *  rotation (transformation). The active rotation will rotate vector r
 *  into r' while the coordinate system K will be unchanged. Or in other
 *  words the original vector r and the resulted vector r' refere to
 *  the same coordinate system K. The passive rotation correspond to
 *  a rotation of the coordinate system K into K' while the point P is
 *  fixed. 
 *  
 *  The rotation matrix can be specified by just three independent
 *  parameters. For more than three parameters constraints exist
 *  between the parameters. There are six constraints on the nine matrix
 *  elements:
 *
 *
 *  The parametrisation using three Euler angles
 *  (a) rotation about fixed (static or non-moving frame)
 *      or rotated (moving frame) axis and
 *
 *      R() = Rz(k) Ry(p) Rx(o)
 *

 *
 *  R = R'x() R'y() R'z()
 *  (b) the sequence of principle axis.
 *
 *   about all 3 axis or by repyted rotation about the 
 *   1st axis.
 *
 *   For the rotation about fixed axies 6 combinations exist:
 *   xyz, zxy, yzx, xzy, yxz, zyx.
 *   Also 6 combinations exist for rotated axies 12 combinations
 *   can be constructed. And for the sequences of coordinates axes
 *   12 possible sequences exist: 
 
 xyz, xyx, yzx, yzy, zxy, zxz,
 *   xzy, xzx, yxz, yxy, zyx and zyz. So in total 24 possibilities
 *   exist. Due to the fact that the resulted rotation matrix
 *   is the same for rotations about fixed axis and rotated axis
 *   in the opposite order
 *
 *       Rxyz(c,b,a) == Rx'y'z'(a,b,c)
 *
 *  the number of can be reduced to 12 combinations.
 *  For the 24 angle set conventions see [Craig 86] Appendix B.
 *  [Shoemake 94] show source code to convert any of this combinations
 *  of Euler angles to and from matrices and quaternions.
 *
 *  Some notes are missing:
 *    - on parametrisation, ambiguity and singularity
 *    - on rotation order, fixed (non-moving reference frame)
 *      and rotated/displaced axes
 *    - on rotation angles and rotation direction
 *
 * Clear naming conventions for rotation angle sets are missing! 
 *
 * ATTENTION:
 *   For angular representation:
 *    - The active rotation in counter-clockwise direction is equal
 *      to passive rotation in clockwise direction, respectivly
 *      active rotation in in clockwise direction is equal
 *      to passive rotation in counter/clockwise direction:
 *
 *         active/ccw == passive/cw
 *         active/cw  == passive/ccw
 *
 *    - The rotation matrix using the sequence kappa-phi-omega gives
 *      R(k)*R(p)*R(o) = (R^t(o)*R^t(p)*R^t(k))^t = [ R(-o,-p,-k) ]^t.
 *
 *  Function calls:
 *  The set-functions will compose the rotation matrix from the given
 *  parameters while the get-functions will decompose/derive the parameters
 *  from the given rotation matrix.
 *
 *  References:
 *  - Books
 *    [Craig 86] Introduction to robotics: mechanics and control,
 *      2. Ed., Addision-Weslwy Publishing Corp., 1986.
 *    [MoP 80] American Society of Photogrammetry: Manual of Photogrammetry,
 *      4th Ed., American Society of Photogrammetry, 1980.
 *    [Finsterwalder/Hofmann 68] Finsterwalder, R., Hofmann, W.:
 *      Photogrammetrie, 3.Aufl, Walter de Gruyter & Co, Berlin, 1968.
 *    [Goldstein 63] Goldstein: Klassische Mechanik, Frankfurt a.M., 1963.
 *    [Konecny/Lehmann 84] Konecny, G., Lehmann, G.: Photogrammetrie, 4.Aufl.,
 *      Walter de Gruyter, Berlin/New York, 1984.
 *    [Kraus 94] Kraus, K.: Photogrammetrie, Band 1, 5.Aufl.,
 *      Duemmler Verlag Bonn, 1994.
 *    [Mikhail 01] Mikhail, E.M., Bethel, J.S., McGlone, J.C.: Introduction
 *      to Modern Photogrammetry, John Wiley & Sons, Inc., 2001.
 *    [Rueger 78] Rueger/Pietschner/Regensburger: Photogrammetrie, 4. Auflage,
 *      1978.
 *    [Schwidefsky/Ackermann 76] Schwidefsky, K., Ackermann, F.:
 *      Photogrammetrie, 7.Aufl, B.G. Teubner, Stuttgart, 1976.
 *    [Thompson 69] Thompson, E.H.: Introduction to the algebra of matrices
 *      with some applications, Adam Hilger LTD, London, 1969.
 *    [Zurmuehl 50] Zurmuehl, R.: Matrizen, Springer-Verlag,
 *      Berlin/Goettingen/Heidelberg, 1950.
 *
 *  - Articles
 *    [Rosenfield 59] Rosenfield, G.H., PE, 25(4), pp. 536-553, 1959.
 *    [Sanso 73] Sanso, F., Photogrammetria, 29, pp. 203-216, 1973.
 *    [Shih 90] Shih, T.-Y., PE, 56(8), pp. 1173-1179, 1990.
 *    [Shoemake 94] Shoemake, Ken,: Euler Angle Conversion, Graphic Gems IV,
 *      Paul S. Heckbert (Ed.), Academic Press, 1994.
 *    [Shut 58] Shut, G.H., Photogrammetria, 15(4), pp. 149-162, 1958.
 *
 *  - Proceedings
 *
 *  - Reports
 *    [Rinner 57] Rinner, K.: Ueber raeumliche Drehungen, DGK Reihe A Nr. 25.
 *
 *  - Others
 *    [FM-I] Flugmechanik I Skript, WS 2000/2001, Institut für Luft-
 *      und Raumfahrt, TU Berlin. 
 *
 *  @author Dirk Stallmann
 */
template<class T>
class RotMatrix3
{
public:
	friend istream& operator>> FRIEND_FUNC_TEMPLATE (istream& is,
		RotMatrix3<T>& M);
	friend ostream& operator<< FRIEND_FUNC_TEMPLATE (ostream& os,
		const RotMatrix3<T>& M);
	friend RotMatrix3<T> operator* FRIEND_FUNC_TEMPLATE (const RotMatrix3<T>& A,
		const RotMatrix3<T>& B);

	/// set up the identity matrix
	void identity();

    /// default constructor, generates a 3x3 identity matrix
	RotMatrix3() { identity(); }

	/// copy constructor
	RotMatrix3(const RotMatrix3& M);

	/// copy assignment operator
	RotMatrix3& operator=(const RotMatrix3& M);

	/// subscript operator using zero-based indexing
	T& operator ()(int i, int j)
	{
		assert((i >= 0) && (i < 3));
		assert((j >= 0) && (j < 3));
		return _a[i][j];
	}

	/// subscript operator using zero-based indexing
	const T& operator ()(int i, int j) const
	{
		assert((i >= 0) && (i < 3));
		assert((j >= 0) && (j < 3));
		return _a[i][j];
	}

	/** Construct the orthogonal matrix from rotation around a single axis.
	 *  The resulted matrix A will carry a point/vector r(x,y,z) into
	 *  r'(x',y',z') by the operation r' = A r. Matrix A represents rotations
	 *  about the fixed axes x, y and z respectively.
	 *
	 * References: [Thompson 69]
	 *
	 * @param axis  rotation around the x-axis ('x'), y-axis ('y')
	 *              or z-axis ('z').
	 * @param a     rotation angle in radian. Positive angles (a > 0) means
	 *              counter-clockwise rotation while negative angles
	 *              (a < 0) means clockwise rotation.
	 */
	void set(char axis, T a);

	/**@name Parametrisation using direction cosines.
	 * In fact the direction cosines are the matrix elements itself.
	 * The predefined matrix elements are copied from or into an array.
	 * The function requires the values of the direction cosines in
	 * row-oriented order, e.g. a11 = v(0), a12 = v(1), ... a33 = v(9).
	 */
	//@{
		/// copy matrix elements from arry
		void set_matrix(const T* v)
		{
			assert(v != 0);
			
			A11 = v[0]; A12 = v[1]; A13 = v[2];
			A21 = v[3]; A22 = v[4]; A23 = v[5];
			A31 = v[6]; A32 = v[7]; A33 = v[8];
		}

		/// copy matrix elements into array
		void get_matrix(T* v) const
		{
			assert(v != 0);

			v[0] = A11; v[1] = A12; v[2] = A13;
			v[3] = A21; v[4] = A22; v[5] = A23;
			v[6] = A31; v[7] = A32; v[8] = A33;
		}
	//@}
	
	/**@name Parametrisation using 3 rotation angles.
	 *  Sequence of rotations : R(o,p,k) = Rz(k) * Ry(p) * Rx(o)
	 *  Coordinate axes       : fixed
	 *  Direction of rotation : clockwise
	 *
	 *  References:
	 *    [MoP 80] p. 51, Gl. 2.23
	 *    [Konecny/Lehmann 84] S. 101 Matrix A
	 *    inpho Stuttgart, Rotation matrix in PATB
	 * ?  [FM-I] Gl. 4.08a, (alte Luftfahrtnorm LN 9300), counter-clockwise!
	 *    See also personal notes!
	 */
	//@{
		/// set matrix using rotation angles
		void set_angles_opk_fix(T omega, T phi, T kappa)
		{
			T so = (T)sin((double)omega);
			T sp = (T)sin((double)phi);
			T sk = (T)sin((double)kappa);

			T co = (T)cos((double)omega);
			T cp = (T)cos((double)phi);
			T ck = (T)cos((double)kappa);

			A11 =  cp * ck;
			A12 =  co * sk + so * sp * ck;
			A13 =  so * sk - co * sp * ck;

			A21 = -cp * sk;
			A22 =  co * ck - so * sp * sk;
			A23 =  so * ck + co * sp * sk;

			A31 =  sp;
			A32 = -so * cp;
			A33 =  co * cp;
		}

		/// get rotation angles from matrix
		void get_angles_opk_fix(T& omega, T& phi, T& kappa) const
		{
			omega = (T)atan2(-A32, A33);
			phi = (T)asin(A31);
			kappa = (T)atan2(-A21, A11);
		}
	//@}
	
	/**@name Parametrisation using 3 rotation angles.
	 *  Sequence of rotations : R(o,p,k) = Rz(k) * Ry(p) * Rx(o)
	 *  Coordinate axes       : fixed
	 *  Direction of rotation : counter-clockwise
	 *
	 *  References:
	 *    [Schwidefsky/Ackermann 76] Tab. 27.1, Spalte 2
	 *    [Albertz/Kreiling 75] S. 217, Drehung um feste Achsen
	 */
	//@{
		/// set matrix using rotation angles
		void set_angles_opk_fix_ccw(T omega, T phi, T kappa)
		{
			T so = (T)sin((double)omega);
			T sp = (T)sin((double)phi);
			T sk = (T)sin((double)kappa);

			T co = (T)cos((double)omega);
			T cp = (T)cos((double)phi);
			T ck = (T)cos((double)kappa);
			
			A11 =  ck * cp;
			A12 = -sk * co + ck * sp * so;
			A13 =  sk * so + ck * sp * co;
			
			A21 =  sk * cp;
			A22 =  ck * co + sk * sp * so;
			A23 = -ck * so + sk * sp * co;
			
			A31 = -sp;
			A32 =  cp * so;
			A33 =  cp * co;
		}

		/// get rotation angles from matrix
		void get_angles_opk_fix_ccw(T& omega, T& phi, T& kappa) const
		{
			omega = (T)atan2(A32, A33);
			phi = (T)asin(-A31);
			kappa = (T)atan2(A21, A11);
		}
	//@}

	/**@name Parametrisation using 3 rotation angles.
	 *  Sequence of rotations : R(p,o,k) = Rz(k) * Rx(o) * Ry(p)
	 *  Coordinate axis       : fixed
	 *  Direction of rotation : counter-clockwise
	 *
	 *  References:
	 *    [Finsterwalder/Hofmann 68] S. 32, Gl. 1.21
	 *    - personal notes.
	 */
	//@{
		/// set matrix using rotation angles
		void set_angles_pok_fix(T phi, T omega, T kappa)
		{
			T sp = (T)sin((double)phi);
			T so = (T)sin((double)omega);
			T sk = (T)sin((double)kappa);

			T cp = (T)cos((double)phi);
			T co = (T)cos((double)omega);
			T ck = (T)cos((double)kappa);

			A11 =  cp * ck - sp * so * sk;
			A12 = -co * sk;
			A13 =  sp * ck + cp * so * sk;

			A21 = cp * sk + sp * so * ck;
			A22 = co * ck;
			A23 = sp * sk - cp * so * ck;

			A31 = -sp * co;
			A32 =  so;
			A33 =  cp * co;
		}

		/// get rotation angles from matrix
		void get_angles_pok_fix(T& phi, T& omega, T& kappa) const
		{
			phi = (T)atan2(-A31, A33);
			omega = (T)asin(A32);
			kappa = (T)atan2(-A12, A22);
		}
	//@}
	
	/**@name Parametrisation using 3 rotation angles.
	 *  Sequence of rotation  : R(o,p,k) = Rx(o) * Ry(p) * Rz(k)
	 *  Coordinate axes       : rotated
	 *  Direction of rotation : counter-clockwise
	 *
	 *  References:
	 *    [Kraus 94] Gl. 2.2-4 oder 2.1-1-6
	 *    [Schwidefsky/Ackermann 76] Gl. 1.7c (1.6e?)
	 *    [Finsterwalder/Hofmann 68] S. 33, Gl. 1.25
	 */
	//@{
		/// set matrix using rotation angles
		void set_angles_opk_rot(T omega, T phi, T kappa)
		{
			T so = (T)sin((double)omega);
			T sp = (T)sin((double)phi);
			T sk = (T)sin((double)kappa);

			T co = (T)cos((double)omega);
			T cp = (T)cos((double)phi);
			T ck = (T)cos((double)kappa);

			A11 =  cp * ck;
			A12 = -cp * sk;
			A13 =  sp;

			A21 =  co * sk + so * sp * ck;
			A22 =  co * ck - so * sp * sk;
			A23 = -so * cp;

			A31 =  so * sk - co * sp * ck;
			A32 =  so * ck + co * sp * sk;
			A33 =  co * cp;
		}

		/// get rotation angles from matrix
		void get_angles_opk_rot(T& omega, T& phi, T& kappa) const
		{
			omega = (T)atan2(-A23, A33);
			phi = (T)asin(A13);
			kappa = (T)atan2(-A12, A11);
		}
	//@}
	
	/**@name Parametrisation using 3 rotation angles.
	 *  Sequence of rotations : R(p,o,k) = Ry(p) * Rx(o) * Rz(k)
	 *  Coordinate axis       : rotated
	 *  Direction of rotation : counter-clockwise
	 *
	 *  References:
	 *    [Kraus 94] S. 380, Gl. 2.2-1-7
	 *    [Schwidefsky/Ackermann 76] Tab. 27.1, Spalte 1
	 *    [Finsterwalder/Hofmann 68] S. 32, Gl. 1.23
	 *    [Konecny/Lehmann 84] S. 101 Matrix B
	 *    [Rueger 78] S. 21, (2.16a)
	 */
	//@{
		/// set matrix using rotation angles
		void set_angles_pok_rot(T phi, T omega, T kappa)
		{
			T sp = (T)sin((double)phi);
			T so = (T)sin((double)omega);
			T sk = (T)sin((double)kappa);

			T cp = (T)cos((double)phi);
			T co = (T)cos((double)omega);
			T ck = (T)cos((double)kappa);

			A11 =  cp * ck + sp * so * sk;
			A12 = -cp * sk + sp * so * ck;
			A13 =  sp * co;

			A21 =  co * sk;
			A22 =  co * ck;
			A23 = -so;

			A31 = -sp * ck + cp * so * sk;
			A32 =  sp * sk + cp * so * ck;
			A33 =  cp * co;
		}

		/// get rotation angles from matrix
		void get_angles_pok_rot(T& phi, T& omega, T& kappa) const
		{
			phi = (T)atan2(A13, A33);
			omega = (T)asin(-A23);
			kappa = (T)atan2(A21, A22);
		}
	//@}

	/**@name Parametrisation using 3 rotation angles.
	 *  Sequence of rotations : R(a,e,r) = Rz(r) * Rx(e) * Rz(a)
	 *  Coordinate axis       : fixed
	 *  Direction of rotation : ?
	 *
	 * This is a special parametrisation implemented in Australis!?!
	 *
	 *  References:
	 *    - Australis source code. (elsewhere?)
	 *    - personal notes.
	 */
	//@{
		/// set matrix using rotation angles
		void set_angles_australis(T az, T elev, T roll)
		{
			T sa = (T)sin((double)az);
			T se = (T)sin((double)elev);
			T sr = (T)sin((double)roll);

			T ca = (T)cos((double)az);
			T ce = (T)cos((double)elev);
			T cr = (T)cos((double)roll);

			A11 =  ca * cr + sa * se * sr;
			A12 = -ca * sr + sa * se * cr;
			A13 =  sa * ce;

			A21 =  sa * cr - ca * se * sr;
			A22 = -sa * sr - ca * se * cr;
			A23 = -ca * ce;

			A31 =  ce * sr;
			A32 =  ce * cr;
			A33 = -se;
		}

		/// get rotation angles from matrix
		void get_angles_australis(T& az, T& elev, T& roll) const
		{
			az = (T)atan2(A13, -A23);
			elev = -(T)asin(A33);
			roll = (T)atan2(A31, A32);
		}
	
	/**@name Parametrisation using small rotation angles.
	 *  The small angles eps_x, eps_y, eps_z are about the respective
	 *  coordinate axes in clockwise direction. "Small" means a few
	 *  angular seconds. The sequence of rotations can be in any order
	 * (small rotations are commutative!).
	 *
	 *  References: [Thompson 69]
	 */
	//@{
		/// set matrix using small rotations
		void set_angles_small(T eps_x, T eps_y, T eps_z)
		{
			A11 = (T)1.;
			A12 = -eps_z;
			A13 =  eps_y;

			A21 =  eps_z;
			A22 = (T)1.;
			A23 = -eps_x;

			A31 = -eps_y;
			A32 =  eps_x;
			A33 = (T)1.;
		}

		/// get small rotations from matrix
		void get_angles_small(T& eps_x, T& eps_y, T& eps_z) const
		{
			eps_x = A32;
			eps_y = A13;
			eps_z = A21;
		}
	//@}

	/**@name Parametrisation using three upper-diagonal matrix elements.
	 *  Applying the theorems of orthogonal matrices six matrix elements
	 *  (a11, a21, a22, a31, a32, a33) can be constructed from three
	 *  off-diagonal elements (a12, a13, a23).
	 *
	 *  References: [Shut 58]
	 */
	//@{
		/// set matrix using three upper-diagonal matrix elements
		void set_upper_elem(T a12, T a13, T a23)
		{
			A12 = a12;
			A13 = a13;
			A23 = a23;

			A11 = (T)sqrt(1. - a12 * a12 - a13 * a13);
			A33 = (T)sqrt(1. - a13 * a13 - a23 * a23);

			T d = 1. - a13 * a13;

			A21 = -(a12 * A33 + A11 * a13 * a23) / d;
			A22 =  (A11 * A33 - a12 * a13 * a23) / d;
			A31 =  a12 * a23 - a13 * A22;
			A32 =  a13 * A21 - A11 * a23;
		}

		/// get upper-diagonal matrix elements from matrix
		void get_upper_elem(T& a12, T& a13, T& a23) const
		{
			a12 = A12;
			a13 = A13;
			a23 = A23;
		}
	//@}

	/**@name Parametrisation by Rodrigues.
	 *  Rodrigues construct the orthogonal matrix from Cayley's formular
	 *
	 *    R = (I - S)(I + S)^(-1)
	 *
	 *  using the skew-symmetric matrix
	 *  
	 *            (  0  -c   b )
	 *    S = 1/2 (  c   0  -a ).
	 *            ( -b   a   0 )
	 *
	 *  References: [Thompson 69]
	 */
	//@{
		/// set matrix Rodrigues parameters
		void set_rodrigues_matrix(T a, T b, T c)
		{
			T aa = a * a;
			T bb = b * b;
			T cc = c * c;

			T d = (aa + bb + cc) / 4. + 1.;

			A11 = (1. + ( aa - bb - cc) / 4.) / d;
			A12 = ((a * b) / 2. - c) / d;
			A13 = ((a * c) / 2. + b) / d;

			A21 = ((a * b) / 2. + c) / d;
			A22 = (1. + (-aa + bb - cc) / 4.) / d;
			A23 = ((b * c) / 2. - a) / d;

			A31 = ((a * c) / 2. - b) / d;
			A32 = ((b * c) / 2. + a) / d;
			A33 = (1. + (-aa - bb + cc) / 4.) / d;
		}

		/// get Rodrigues parameters from matrix
		void get_rodrigues_matrix(T& a, T& b, T& c) const
		{
			T d2 = 2. / (A11 + A22 + A33 + 1);

			a = (A32 - A23) * d2;
			b = (A13 - A31) * d2;
			c = (A21 - A12) * d2;
		}
	//@}

	/**@name (Hamiltonian) unit quaternion representation.
	 *  The transformation can be also expressed by means of
	 *  quaternion algebra
	 *
	 *    r' = q r q^(-1)
	 *
	 *  where
	 *    q = (q0, q1, q2, q3) is a unit quaternion and
	 *    q^(-1) is the recipropcal of q.
	 *
	 *  By applying quaternion multiplication the rotation matrix
	 *  can be derived.
	 *
	 *  References:
	 *    [Shut 58]
	 *    [Sanso 73]
	 *    [Mikhail 01] p. 452, (E-39)
	 *    [Craig 86] p. 55, (2.91) - (2.92) ???
	 */
	//@{
		/// set matrix from unit quaternion
		void set_quaternion(T q0, T q1, T q2, T q3)
		{
			T q0q0, q0q1, q0q2, q0q3,
					q1q1, q1q2, q1q3,
						  q2q2, q2q3,
								q3q3;

			q0q0 = q0 * q0;
			q0q1 = q0 * q1;
			q0q2 = q0 * q2;
			q0q3 = q0 * q3;
			q1q1 = q1 * q1;
			q1q2 = q1 * q2;
			q1q3 = q1 * q3;
			q2q2 = q2 * q2;
			q2q3 = q2 * q3;
			q3q3 = q3 * q3;

			A11 = q0q0 + q1q1 - q2q2 - q3q3;
			A12 = 2.0 * (q1q2 - q0q3);
			A13 = 2.0 * (q1q3 + q0q2);
			A21 = 2.0 * (q1q2 + q0q3);
			A22 = q0q0 - q1q1 + q2q2 - q3q3;
			A23 = 2.0 * (q2q3 - q0q1);
			A31 = 2.0 * (q1q3 - q0q2);
			A32 = 2.0 * (q2q3 + q0q1);
			A33 = q0q0 - q1q1 - q2q2 + q3q3;
		}

		/// set matrix from unit quaternion (STL vector)
		void set_quaternion(const vector<T>& q)
		{
			assert(q.size() == 4);
			set_quaternion(q[0], q[1], q[2], q[3]);
		}

		/// get unit quaternion from matrix
		void get_quaternion(T& q0, T& q1, T& q2, T& q3) const
		{
			q0 = 0.5 * (T)sqrt(_a[0][0] + _a[1][1] + _a[2][2] + 1.0);

			if(q0 != 0.0)
				{
				q1 = -(_a[1][2] - _a[2][1]) / q0 / 4.0;
				q2 =  (_a[0][2] - _a[2][0]) / q0 / 4.0;
				q3 = -(_a[0][1] - _a[1][0]) / q0 / 4.0;
				}
			else
				{
				T q1q2, q1q3, q2q3;

				q1q2 = (_a[0][1] + _a[1][0]) / 4.0;
				q1q3 = (_a[0][2] + _a[2][0]) / 4.0;
				q2q3 = (_a[1][2] + _a[2][1]) / 4.0;

				if (q1q2 != 0.0)
					{
					q3 = (T)sqrt(q1q3 * q2q3 / q1q2);
					q1 = q1q3 / q3;
					q2 = q2q3 / q3;
					}
				else if (q1q3 != 0.0)
					{
					q2 = (T)sqrt(q1q2 * q2q3 / q1q3);
					q1 = q1q2 / q2;
					q3 = q2q3 / q2;
					}
				else if (q2q3 != 0.0)
					{
					q1 = (T)sqrt(q1q2 *q1q3 / q2q3);
					q2 = q1q2 / q1;
					q3 = q1q3 / q1;
					}
				else
					{
			// This means there is only the main diagonal left in
			// the rotation matrix! The elements are
			//
			//	( q1q1-q2q2-q3q3          0                0       )
			//	(        0        -q1q1+q2q2-q3q3          0       )
			//	(        0                0        -q1q1-q2q2+q3q3 )
			// Since only one element of the (normalized) quaternion
			// is != 0, it is +/-1 (we choose +1). Depending on which
			// element, we have the following signs in the main diagonal:
			//	q1:	+ - -
			//	q2:	- + -
			//	q3:	- - +
					if (_a[0][0] > 0.0)
						{
						q1 = 1.0; q2 = q3 = 0.0;
						}
					else if (_a[1][1] > 0.0)
						{
						q2 = 1.0; q1 = q3 = 0.0;
						}
					else
						{
						q3 = 1.0; q1 = q2 = 0.0;
						}
					}
				}
		}

		/// get unit quaternion (STL vector) from matrix
		void get_quaternion(vector<T>& q) const
		{
			assert(q.size() == 4);
			get_quaternion(q[0], q[1], q[2], q[3]);
		}
	//@}
	
		/// set matrix from unit quaternion
		void set_quaternion2(T q0, T q1, T q2, T q3)
		{
			T q0q0, q0q1, q0q2, q0q3,
					q1q1, q1q2, q1q3,
						  q2q2, q2q3,
								q3q3;

			q0q0 = q0 * q0;
			q0q1 = q0 * q1;
			q0q2 = q0 * q2;
			q0q3 = q0 * q3;
			q1q1 = q1 * q1;
			q1q2 = q1 * q2;
			q1q3 = q1 * q3;
			q2q2 = q2 * q2;
			q2q3 = q2 * q3;
			q3q3 = q3 * q3;

			A11 = 1.0 - 2.0 *(q2q2 + q3q3);
			A12 = 2.0 * (q1q2 - q0q3);
			A13 = 2.0 * (q1q3 + q0q2);
			A21 = 2.0 * (q1q2 + q0q3);
			A22 = 1.0 - 2.0 * (q1q1 + q3q3);
			A23 = 2.0 * (q2q3 - q0q1);
			A31 = 2.0 * (q1q3 - q0q2);
			A32 = 2.0 * (q2q3 + q0q1);
			A33 = 1.0 - 2.0 * (q1q1 + q2q2);
		}
		
		/// set matrix from unit quaternion (STL vector)
		void set_quaternion2(const vector<T>& q)
		{
			assert(q.size() == 4);
			set_quaternion2(q[0], q[1], q[2], q[3]);
		}
				/// get unit quaternion from matrix
		void get_quaternion2(T& q0, T& q1, T& q2, T& q3) const
		{
			q0 = 0.5 * (T)sqrt(_a[0][0] + _a[1][1] + _a[2][2] + 1.0);

			if(q0 != 0.0)
				{
				q1 = -(_a[1][2] - _a[2][1]) / q0 / 4.0;
				q2 =  (_a[0][2] - _a[2][0]) / q0 / 4.0;
				q3 = -(_a[0][1] - _a[1][0]) / q0 / 4.0;
				}
			else
				{
				T q1q2, q1q3, q2q3;

				q1q2 = (_a[0][1] + _a[1][0]) / 4.0;
				q1q3 = (_a[0][2] + _a[2][0]) / 4.0;
				q2q3 = (_a[1][2] + _a[2][1]) / 4.0;

				if (q1q2 != 0.0)
					{
					q3 = (T)sqrt(q1q3 * q2q3 / q1q2);
					q1 = q1q3 / q3;
					q2 = q2q3 / q3;
					}
				else if (q1q3 != 0.0)
					{
					q2 = (T)sqrt(q1q2 * q2q3 / q1q3);
					q1 = q1q2 / q2;
					q3 = q2q3 / q2;
					}
				else if (q2q3 != 0.0)
					{
					q1 = (T)sqrt(q1q2 *q1q3 / q2q3);
					q2 = q1q2 / q1;
					q3 = q1q3 / q1;
					}
				else
					{
			// This means there is only the main diagonal left in
			// the rotation matrix! The elements are
			//
			//	( q1q1-q2q2-q3q3          0                0       )
			//	(        0        -q1q1+q2q2-q3q3          0       )
			//	(        0                0        -q1q1-q2q2+q3q3 )
			// Since only one element of the (normalized) quaternion
			// is != 0, it is +/-1 (we choose +1). Depending on which
			// element, we have the following signs in the main diagonal:
			//	q1:	+ - -
			//	q2:	- + -
			//	q3:	- - +
					if (_a[0][0] > 0.0)
						{
						q1 = 1.0; q2 = q3 = 0.0;
						}
					else if (_a[1][1] > 0.0)
						{
						q2 = 1.0; q1 = q3 = 0.0;
						}
					else
						{
						q3 = 1.0; q1 = q2 = 0.0;
						}
					}
				}
		}

		/// get unit quaternion (STL vector) from matrix
		void get_quaternion2(vector<T>& q) const
		{
			assert(q.size() == 4);
			get_quaternion2(q[0], q[1], q[2], q[3]);
		}
		
	/**@name Parametrisation by one rotation about a line.
	 *  The rotation is defined by a line in space and an angle t
	 *  about this line. The line is defined by
	 *  the direction cosines a, b and c.
	 *  Also called "equivalent angle-axis representation" (Craig 86).
	 *
	 *  References:
	 *    [Mikhail 01] p. 373, (A-68)
	 *    [Rinner 57] S. 28, (4b)
	 *    [Rueger 78] S. 19, (2.7a) - (2.7c)
	 *    [Craig 86] p. 52, (2.80) - (2.82)
	 */
	//@{
		/// set matrix by parameters
		void set_about_line(T a, T b, T c, T t)
		{
			T st = (T)sin((double)t);
			T ct = (T)cos((double)t);

			T e = 1. - ct;

			A11 = a * a * e +     ct;
			A12 = a * b * e - c * st;
			A13 = a * b * e + b * st;
			
			A21 = a * b * e + c * st;
			A22 = b * b * e +     ct;
			A23 = b * c * e - a * st;
			
			A31 = a * c * e - b * st;
			A32 = b * c * e + a * st;
			A33 = c * c * e +     ct;
		}

		/// get parameters from matrix
		void get_about_line(T& a, T& b, T& c, T& t) const
		{
			t = acos((A11 + A22 + A33 - 1.) / 2.);
			
			T f = 1. / (2. * sin(t));
			
			a = f * (A32 - A23);
			b = f * (A13 - A31);
			c = f * (A21 - A12);
		}
	//@}

	/**@name set/get rotation matrix using different representations.
	 *  @param v     pointer to array of rotation parameters
	 *  @param type  type of parametrisation
	 */
	//@{
		/// set matrix from parameters
		void set(const T* v, ERotMatrix3::ecode type)
		{
			assert(v != 0);
			
			switch (type)
				{
				case ERotMatrix3::UNDEFINED:
					cerr << "RotMatrix3<T>::set: type undefined" << endl;
					assert(false);
					break;

				case ERotMatrix3::ROTATION_MATRIX:
				case ERotMatrix3::DIRECTION_COSINES:
					set_matrix(v);
					break;

				case ERotMatrix3::ANGLES_OPK_FIX:
					set_angles_opk_fix(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_OPK_FIX_CCW:
					set_angles_opk_fix_ccw(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_POK_FIX:
					set_angles_pok_fix(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_OPK_ROT:
					set_angles_opk_rot(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_POK_ROT:
					set_angles_pok_rot(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_AUSTRALIS:
					set_angles_australis(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_SMALL:
					set_angles_small(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::UPPER_DIAG_ELEM:
					set_upper_elem(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::RODRIGUES_MATRIX:
					set_rodrigues_matrix(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::UNIT_QUATERNION:
					set_quaternion(v[0], v[1], v[2], v[3]);
					break;

				case ERotMatrix3::ROTATION_ABOUT_LINE:
					set_about_line(v[0], v[1], v[2], v[3]);
					break;
					
				default:
					cerr << "RotMatrix3<T>::set: bad type: " << type << endl;
					assert(false);
				}
		}
		
		/// get parameters from matrix
		void get(T* v, ERotMatrix3::ecode type) const
		{
			assert(v != 0);
			
			switch (type)
				{
				case ERotMatrix3::UNDEFINED:
					cerr << "RotMatrix3<T>::get: type undefined" << endl;
					assert(false);
					break;

				case ERotMatrix3::ROTATION_MATRIX:
				case ERotMatrix3::DIRECTION_COSINES:
					get_matrix(v);
					break;

				case ERotMatrix3::ANGLES_OPK_FIX:
					get_angles_opk_fix(v[0], v[1], v[2]);
					break;
					
				case ERotMatrix3::ANGLES_OPK_FIX_CCW:
					get_angles_opk_fix_ccw(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_OPK_ROT:
					get_angles_opk_rot(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_POK_ROT:
					get_angles_pok_rot(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_AUSTRALIS:
					get_angles_australis(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::ANGLES_SMALL:
					get_angles_small(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::UPPER_DIAG_ELEM:
					get_upper_elem(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::RODRIGUES_MATRIX:
					get_rodrigues_matrix(v[0], v[1], v[2]);
					break;

				case ERotMatrix3::UNIT_QUATERNION:
					get_quaternion(v[0], v[1], v[2], v[3]);
					break;

				case ERotMatrix3::ROTATION_ABOUT_LINE:
					get_about_line(v[0], v[1], v[2], v[3]);
					break;

				default:
					cerr << "RotMatrix3<T>::get: bad type: " << type << endl;
					assert(false);
				}
		}
	//@}

	/**@name set/get rotation matrix using different representations.
	 *  @param v     vector of rotation parameters
	 *  @param type  type of parametrisation
	 */
	//@{
		/// set matrix from parameters
		void set(const vector<T>& v, ERotMatrix3::ecode type)
		{
			assert(!v.empty());
			
			ERotMatrix3 map;
			assert(v.size() == map.size(type));
			set(&v[0], type);
		}

		/// get parameters from matrix
		void get(vector<T>& v, ERotMatrix3::ecode type) const
		{
			ERotMatrix3 map;
			v.resize(map.size(type));
			get(&v[0], type);
		}
	//@}

	/// transpose matrix 'this'
	void transpose();

	/** invert matrix 'this'.
	 *  Simply a synonym for transpose.
	 */
	void inverse() { transpose(); }

	/** returns the value of the determinate of 'this' matrix.
	 *  Using Regel von Sarrus. The determinate of a orthogonal matrix
	 *  det(A) = +/-1! The determinate is positive for a pure rotation
	 *  and negative for a rotation combined with a reflection.
	 *  Useful for testing if the matrix is orthogonal: A'A = AA' = I.
	 */
	T det() const;

	/** test if all elements in matrix 'this' and the matrix 'A' are equal.
	 *  @param A     second matrix
	 *  @param eps   test limit
	 */
	bool is_equal(RotMatrix3& A, T eps=0) const;

private:
	/// matrix elements
	T _a[3][3];
/*
	// anonymous unions does't work width g++ 2.95.2
	union {
		struct {
			T _11, _12, _13;
			T _21, _22, _23;
			T _31, _32, _33;
		};
		T _a[3][3];
	};
*/
};


// non-member functions...

/** input operator.
 *  Read matrix 'M' from input stream 'is'.
 */
template<class T>
istream& operator>>(istream& is, RotMatrix3<T>& M)
{
	is >> M.A11 >> M.A12 >> M.A13;
	is >> M.A21 >> M.A22 >> M.A23;
	is >> M.A31 >> M.A32 >> M.A33;

	return is;
}


/** output operator.
 *  Print matrix 'M' on the output stream 'os'.
 */
template<class T>
ostream& operator<<(ostream& os, const RotMatrix3<T>& M)
{
	os << M.A11 << " " << M.A12 << " " << M.A13 << endl;
	os << M.A21 << " " << M.A22 << " " << M.A23 << endl;
	os << M.A31 << " " << M.A32 << " " << M.A33 << endl;

	return os;
}


/** matrix product.
 *  Multiply two rotation matrices 'A' and 'B': C = A * B.
 */
template<class T>
RotMatrix3<T> operator*(const RotMatrix3<T>& A, const RotMatrix3<T>& B)
{
	RotMatrix3<T> C;

	C.A11 = A.A11 * B.A11 + A.A12 * B.A21 + A.A13 * B.A31;
	C.A12 = A.A11 * B.A12 + A.A12 * B.A22 + A.A13 * B.A32;
	C.A13 = A.A11 * B.A13 + A.A12 * B.A23 + A.A13 * B.A33;

	C.A21 = A.A21 * B.A11 + A.A22 * B.A21 + A.A23 * B.A31;
	C.A22 = A.A21 * B.A12 + A.A22 * B.A22 + A.A23 * B.A32;
	C.A23 = A.A21 * B.A13 + A.A22 * B.A23 + A.A23 * B.A33;

	C.A31 = A.A31 * B.A11 + A.A32 * B.A21 + A.A33 * B.A31;
	C.A32 = A.A31 * B.A12 + A.A32 * B.A22 + A.A33 * B.A32;
	C.A33 = A.A31 * B.A13 + A.A32 * B.A23 + A.A33 * B.A33;

/*
	for (int i = 0; i < 3; ++i)
		for (int j = 0; j < 3; ++j)
			{
			T sum = (T)0;

			for	(int k = 0; k < 3; ++k)
				sum += A(i,k) * B(k,j);
	
			C(i,j) = sum;
			}
*/

	return C;
}


// implementation

// set up the identity matrix

template<class T>
void RotMatrix3<T>::identity()
{
	A11 = A22 = A33 = T(1);
	A12 = A13 = A21 = A23 = A31 = A32 = T(0);
}


// copy constructor

template<class T>
RotMatrix3<T>::RotMatrix3(const RotMatrix3<T>& M)
{
	memcpy(_a, &M._a, sizeof(_a));
}


// copy assignment

template<class T>
RotMatrix3<T>& RotMatrix3<T>::operator=(const RotMatrix3<T>& M)
{
	memcpy(_a, &M._a, sizeof(_a));
	return *this;
}


// set matrix from rotation around a single axis 

template<class T>
void RotMatrix3<T>::set(char axis, T a)
{
	identity();

	T sin_a = (T)sin((double)a);
	T cos_a = (T)cos((double)a);

	int m = 0;
	switch (axis)
		{
		case 'x': m = 1; break;
		case 'y': m = 2; break;
		case 'z': m = 0; break;
		default:
			cerr << "RotMatrix3<T>::set: bad axis " << axis << endl;
			return;
		}			
	int n = (m + 1) % 3;

	_a[m][m] =  cos_a;
	_a[n][n] =  cos_a;
	_a[m][n] = -sin_a;
	_a[n][m] =  sin_a;
	//_a[m][n] =  sin_a;
	//_a[n][m] = -sin_a;

	return;
}


// transpose rotation matrix

template<class T>
void RotMatrix3<T>::transpose()
{
	T t;

	t = A21; A21 = A12; A12 = t;
	t = A31; A31 = A13; A13 = t;
	t = A32; A32 = A23; A23 = t;
}


// determinate of matrix 'this' (Regel von Sarrus)

template<class T>
T RotMatrix3<T>::det() const
{
	return A11 * A22 * A33 + A12 * A23 * A31 + A13 * A21 * A32 -
		A13 * A22 * A31 - A11 * A23 * A32 - A12 * A21 * A33;
}


// test if all elements are equal

template<class T>
bool RotMatrix3<T>::is_equal(RotMatrix3<T>& A, T eps) const
{
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			if (fabs(_a[i][j] - A._a[i][j]) > eps) return false;

	return true;
}


// non-member functions ...

} // namespace GenLib2

#endif /* GENLIB2_ROTMATRIX3_INCLUDED */
