/**********************************************************************/
//
// FILE:    sstransf.cpp
//          General conversion of rotation parameters.
//          This program convert different types rotation parameters 
//          of the exterior orientation (EO). Each set of EO consist of
//          the following parameters:
//
//            id             - identifier, the  image name or number
//            t              - (GPS-)time (optional)
//            Xo, Yo, Zo     - 3 perspective centre coordinates (optional)
//            p1, p2, ... p9 - 3 until 9 rotations parameters.
//
//          The EO data is read from in_file, the rotation parameters
//          are converted and written to the out_file.
//          The parameters id, time and perspective centre coordinates
//          are handed as strings and therefore unchanged!
//          The input and output format is:
//
//             id   [t]   [Xo   Yo   Zo]   p1   p2   p3  [... p9]
//
//          For output the following the formats are used:
//
//             Parameter        Data typ   Width   Precision   c-format
//
//             id               string      8      -           %8s 
//             t                string     12      -           %12s
//             Xo, Yo, Zo       string     12      -           %12s
//             p1, p3, ... p9   fixed      12      7           %12.7f
//
//          The input and output type is defined by the 1st and 2nd
//          argument. E.g: "opk:deg" as the 1st argument means
//          for input the rotation sequence is omega-phi-kappa (opk)
//          and the unit of rotation angles is degrees (deg).
//          The transponse of the rotation matrix can be done before
//          and/or after the change of signs (CHS)!
//
//          Options:
//            -help        Show this text.
//            -chs_c0      Change sign of column 0.
//            -chs_c1      Change sign of column 1.
//            -chs_c2      Change sign of column 2.
//            -trn_before  Transpose rotation matrix before CHS.
//            -trn_after   Transpose rotation matrix after CHS.
//
//			In general the sequence of operations in this order:
//            1. calculate rotation matrix
//            2. transpose matrix before
//            3. CHS of columns
//            4. transpose matrix after
//            5. calculate rotation angles
//            
// CREATED: 10 Dec 00  (ds)
// CHANGES: 14 Aug 01  Using new rotation matrix class RotMatrix. (ds)
//          17 Aug 01  Using new rotation matrix class RotMatrix3. (ds)
//         	20 Mar 02  Add special azimuth-elevation-roll parametrisation
//                     implemented in Australis. (ds)
// REFS:
//
// (c) 2000-02 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <stdlib.h>
#include <string>
#include <strstream.h>
#include <fstream>
#include <vector>
#include "form.h"
#include "sstransf.h"

const char* const program_name = "sstransf";
const char* const program_version = "version 0.1 (05/03)";

static void usage(int status=0)
{
	cerr << "Usage: " << program_name
		<< " [options] in_type in_unit in_file out_type out_unit out_file\n"
		<< "\n"
		<< "in_type/out_type:\n"
		<< "  rot-matrix     - rotation matrix\n"
		<< "  angles-opk-rot - rotation angles, omega-phi-kappa, rotated axes\n"
		<< "  angles-pok-rot - rotation angles, phi-omega-kappa, rotated axes\n"
		<< "  angles-opk-fix - rotation angles, omega-phi-kappa, fixed axes\n"
		<< "\n"
		<< "in_unit/out_unit:\n"
		<< "  rad - radian\n"
		<< "  deg - degree\n"
		<< "  gon - gon/grad\n"
		<< "\n"
		<< "in_file/out_file: I/O files with data format for each line:\n"
		<< "ID [t] [Xo Yo Zo] p1 ... pn\n"
		<< "  ID           - point ID\n"
		<< "  t            - GPS time (optional)\n"
		<< "  Xo, Yo, Zo   - perspective centre coords. (optional)\n"
		<< "  p1, ... pn   - rotation parameters.\n"
		<< "\n"
		<< "Options:\n"
		<< "  -help             Show this text.\n"
		<< "  -chs_c0           CHS of column 0.\n"
		<< "  -chs_c1           CHS of column 1.\n"
		<< "  -chs_c2           CHS of column 2.\n"
	//"  -hdr id name zf   Write DEP header.\n"
	//"                      id = EO identifier.\n"
	//"                      name = EO name.\n"
	//"                      zf = scan frequency.\n"
		<< "  -pcc              Has perspective centre coords. (Xo, Yo, Zo).\n"
		<< "  -time             Has GPS time.\n"
		<< "  -trn_before       Transpose rotation matrix before CHS.\n"
		<< "  -trn_after        Transpose rotation matrix.after CHS\n"
//		<< "  -skip n           Skip n leading rows.\n";
		<< "  -version          Show program version.\n";
	exit(status);
};


static void error(const char *s, const char *s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}


int main(int argc, char *argv[])
{
	using namespace GenLib2;

	char*		in_fname = NULL;	// input filename
	char*		out_fname = NULL;	// output filename
	int			verb = 0;

	if (argc == 1) usage(1);

	for (int i = 1, j = 0; i < argc; i++)
		{
		if (**(argv + i) == '-')
			{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-verbose") == 0)
				verb = 1;
			else if (strcmp(argv[i], "-version") == 0)
				{
				cerr << program_name << " " << program_version
					<< " compiled " << __DATE__ << " " << __TIME__ << endl;
				exit(0);
				}
			else
				error("Illegal option:", argv[i]);
			}
		else
			{
			switch (j++)
				{
				case 0:
					in_fname = argv[i];
					break;

				case 1:
					out_fname = argv[i];
					break;

				default:
					error("Too many arguments given:", argv[i]);
				}
			}
		}

	ifstream is(in_fname);
	if (!is) error("Cannot open input file:", in_fname);

	ofstream os(out_fname);
	if (!os) error("Cannot open output file:", out_fname);

	SSTransform trf;

	trf.read(is);

	//if (verbose)
	trf.print(cerr);

	Form	str8; str8.width(8);
	Form	fix(3); fix.fixed().width(12);

	string id;
	double x1, y1, z1;
	double x2, y2, z2;

	//read ID
	//while (is >> id >> t >> xo >> yo >> zo)
	while (is >> id)
		{
		is >> x1 >> y1 >> z1;

		trf.transform(x1, y1, z1, x2, y2, z2);

		os << str8(id) << " " << fix(x2) << " " << fix(y2) << " " << fix(z2)
			<< endl;
		}

	return 0;
}
